/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/*
 * $Id: mmap.c,v 1.41.4.6 2002/12/03 14:02:50 gjertsen Exp $
 *
 * $Log: mmap.c,v $
 * Revision 1.41.4.6  2002/12/03 14:02:50  gjertsen
 * On Linux, disable_lock returns output of local_irq_save, which is a long
 * (some machine state register contents, usually). We were truncating it to
 * an int, which is problematic on 64-bit architectures, so introduce new type,
 * cxiIntrState_t for storing disable_lock output.
 *
 * Revision 1.41.4.5  2002/11/05 05:56:01  mcnabb
 * Cleanup:  Free page request list does not need to be doubly linked.  Don't
 * wake up pager kproc when waiting for free buf entry since nothing was added
 * to queue.  Don't wake up FreeEventWord unless a thread is waiting.
 *
 * Revision 1.41.4.4  2002/11/04 18:56:55  mcnabb
 * Fix kernel seg faults caused by putting buf structs that were allocated
 * from the stack onto the free buf list.  This happened if a B_WRITE request
 * got an error.
 *
 * Revision 1.41.4.3  2002/10/29 07:19:58  mcnabb
 * Modifications for RedHat Advanced Server 2.1.
 *
 * Revision 1.41.4.2  2002/09/24 23:33:30  mcnabb
 * Fix Linux multnode mmap
 *
 * Revision 1.41.4.1  2002/09/06 15:48:07  mcnabb
 * Updated #ifdef to make use of new SUSE_LINUX define
 *
 * Modify ifdefs for kernel patches.
 *
 * ifdef'd in pte_kunmap needed for SuSE 2.4.18 kernel
 *
 * Revision 1.41  2002/08/07 19:09:26  radhak
 * Node crashed in flush_vma while getting page table entry from bad pmd
 * offset. This is an old bug. But, for some reason we did not catch it
 * before. Make sure that the pmd is valid before getting page table entry.
 *
 * Revision 1.40  2002/08/06 21:47:19  dcraft
 * Cut down on the number of kernel levels that are irrelevent to define
 *
 * Revision 1.39  2002/07/31 20:54:49  gjertsen
 * Only include Linux mmap patch if MMAP_LINUX_PATCH is toggled on in
 * LINUX_PATCH_DEFINES.
 *
 * Revision 1.38  2002/07/31 14:30:52  radhak
 * Modification to make mmap work on 2.4.18-3 kernel.
 *
 * Revision 1.37  2002/07/30 20:01:48  radhak
 * Linux 2.4.18-5: Before clearing pte entry, call page_remove_rmap() to
 * remove pte entry from page pte chain.
 *
 * Revision 1.36  2002/07/12 15:57:00  radhak
 * After calling filemap_fdatasync(), need to wait until all dirty mmaped
 * pages are flushed disk.
 *
 * Revision 1.35  2002/07/08 23:17:13  tee
 * Release mmap lock before doing mmap I/O.
 *
 * Revision 1.34  2002/06/27 03:08:04  tee
 * Since GPFS data buffers no longer have to be kept for pages that are mapped
 * read-only, do not mark buffer descriptor as having mapped pages unless the
 * pages are writeable.
 *
 * Revision 1.33  2002/06/27 00:23:51  tee
 * If a file is mapped read-only, do not invalidate pages in VMM cache when
 * GPFS data buffer is stolen or the file is closed.  Wait until byte-range
 * token goes away to invalidate.  GPFS data buffer is still required for
 * all writeable mapped pages.
 *
 * Revision 1.32  2002/05/30 23:37:07  dcraft
 * Make these slightly more consistent between emperor and lyndhurst.
 * No functional change
 *
 * Revision 1.31  2002/05/17 22:05:46  dcraft
 * Changes to allow GPFS to run on RH 7.3 (2.4.18-03 kernel).
 * Works at least for 2 level pft
 *
 * Revision 1.30  2002/04/16 18:39:25  wyllie
 * Improve some traces
 *
 * Revision 1.29  2002/03/15 22:28:01  gjertsen
 * Change back to conservative TLB flushing on IA64 for now.
 *
 * Revision 1.28  2002/02/01 19:55:44  gjertsen
 * Quickie fix for ia64 tlb flush problem in mmap. Better fix coming soon.
 *
 * Revision 1.27  2002/01/24 15:20:54  radhak
 * Defect 357492:
 * Flush dirty pages before terminating mmap.
 *
 * Revision 1.26  2001/11/20 20:42:54  radhak
 * Defect 353917:
 * vm_ops->close() always be done after calling vm_ops->open() even if
 * returns error. Because of this Linux semantics we need to increment
 * mmap counters even if an error occurs.
 *
 * Revision 1.25  2001/10/30 16:24:10  gjertsen
 * Add macros for flush tlb operations. Optimize IA64 code to more
 * selectively flush the tlb cache.
 *
 * Revision 1.24  2001/10/03 14:46:14  dcraft
 * First attempt to bring us up to 2.4.9 and 2.4.10
 *
 * Revision 1.23  2001/09/25 19:02:38  gjertsen
 * Some IA64 code cleanup. Suppress unwanted compiler warnings.
 *
 * Revision 1.22  2001/09/25 14:27:51  dcraft
 * Comment change.  No code change.
 *
 * Revision 1.21  2001/09/24 13:15:14  radhak
 * Got rid of dead code.
 *
 * Revision 1.20  2001/08/21 01:56:27  tee
 * Make AIX and Linux versions of buf queue handling be the same.
 *
 * Revision 1.19  2001/08/20 15:02:16  gjertsen
 * Avoid calling flush_tlb_page for IA64 for now (not exported by kernel).
 *
 * Revision 1.18  2001/08/18 02:11:30  tee
 * Improve matching between AIX and Linux mmap code to prepare for source file
 * merge.
 *
 * Revision 1.17  2001/08/16 01:26:24  tee
 * Fix dynamic mmap kproc starting and stopping on Linux.  Don't start processes
 * that only live for the duration of a page fault.  Use the daemon startup
 * thread as a master controller for the kprocs, which will wait on them when
 * they exit to keep them from becoming zombies.
 *
 * Revision 1.16  2001/08/14 02:53:42  tee
 * Don't dereference NULL vinfo pointer on write requests.  Return error to
 * page fault hander if read fails rather than allowing application to use
 * whatever happens to be in the buffer.  Don't clear PageUptodate flag on
 * write requests.  Don't use PageError flag.
 *
 * Revision 1.15  2001/08/13 20:40:51  schmuck
 * Partial fix for Defect 345614 (Raleigh Defect 4423):
 *   The sendfile() system call generates paging requests without an explicit
 * mmap().  When the file is closed, paging requests for read-ahead that were
 * generated as a side effect of sendfile() may still be queued and were not
 * being cleaned up.  A pager kproc would then process these requests using a
 * pointer to the old vinfo structure, which could have been reused for a
 * different file.  This caused inconsistencies in our prefetch buffer
 * accounting, eventually leading to the observed asserts.
 *   Add a flag to the vinfo structure to remember whether we received paging
 * requests for this instance.  On close, check the flag and call mmFlushseg
 * to cancel pending requests if necessary.
 *   Still missing: close needs to wait if a paging request from the instance
 * is currently being processed by a pager kproc.
 *
 * Revision 1.14  2001/08/10 22:33:03  tee
 * Declare functions with correct argument types to avoid unnecessary casting.
 * Cosmetic changes to make AIX and Linux mmap code more similar.
 *
 * Revision 1.13  2001/08/10 22:31:09  tee
 * Don't trace mmap buffer as string since it could contain unprintable
 * characters and might not be terminated.
 *
 * Revision 1.12  2001/08/10 18:22:32  gjertsen
 * Put errno predeclaration for kernel code in Shark-gpl.h.
 *
 * Revision 1.11  2001/08/09 21:11:22  dcraft
 * Modifications to allow running on latest Redhat 7.1 update
 * Kernel version 2.4.3-12.
 * Requires checkout of new site.mcr.proto
 *
 * Revision 1.10  2001/08/04 00:42:27  tee
 * Remove LINUX_MMAP ifdefs
 *
 * Revision 1.9  2001/08/01 20:20:29  tee
 * Increase trace level in flush_vma
 *
 * Revision 1.8  2001/07/21 00:30:55  tee
 * Invalidate memory mapped pages when a byte-range write lock is acquired,
 * unless it is being acquired to service a page fault.  This fixes the
 * problem of stale data being seen sometimes by an mmap application even
 * after a sync.  Also, add new option to kxMmapFlush to discard data without
 * flushing.  Use this in trunc to invalidate mapped pages, since flushing is
 * not necessary for pages in the truncated region.
 *
 * Revision 1.7  2001/07/19 17:39:06  manoj
 * Allow readpage to service sendfile(). Not complete (need to invalidate Linux's
 * cached inode pages, either on return from readpage or on a wx/sx BRL request).
 *
 * Revision 1.6  2001/05/22 13:59:00  radhak
 * Initialize start address before invalidating mmaped pages.
 *
 * Revision 1.5  2001/05/18 19:24:00  radhak
 * Defect 339937:
 * Enabled single node mmap semantics and fixed deadlock found by Tom
 *
 * Revision 1.4  2001/05/08 13:40:57  dixonbp
 * Release the file table inherrited from mmfsd at the top of
 * kernel threads (in pagerKprocMainThreadP).
 *
 * Revision 1.3  2001/05/04 13:41:27  radhak
 * Defect 338936: cont.
 * Moved __WCLONE flag mmap.c
 *
 * Revision 1.2  2001/05/03 20:50:42  radhak
 * Defect 338936:
 * Linux mmap: PagerKproc::adjustKprocs:
 * Need to get rid of kernel thread zombies
 *
 * Revision 1.1  2001/04/02 14:33:16  dixonbp
 * Convert mmap.C to mmap.c
 *
 * Revision 1.18  2001/03/01 20:44:23  radhak
 * Need serialization between nopage and mmap flush.
 * Also, always get page table lock while holding page lock.
 *
 * Revision 1.15  2001/01/09 19:14:45  radhak
 * LINUX_MMAP: enable waitForFreeEntry()
 *
 * Revision 1.14  2000/12/29 22:22:35  radhak
 * Defect 322452: Before calling gpfs_filemap_sync get lock.
 * Also added some traces.
 *
 * Revision 1.13  2000/12/19 16:10:28  wyllie
 * Move Linux mmap declarations out of platform-independent code
 *
 * Revision 1.12  2000/12/15 13:56:46  gjertsen
 * Clean up documentation.
 *
 * Revision 1.11  2000/12/12 17:46:15  wyllie
 * Run cindent, add spaces after commas.  No functional changes.
 *
 * Revision 1.10  2000/12/03 01:48:10  radhak
 * LINUX_MMAP: mmap flush
 *
 * Revision 1.9  2000/12/01 02:10:59  schmuck
 * Instead of assigning NULL function pointers when initializing or resetting the
 * gpfs_operations table, have it point to a dummy function that returns ENOSYS.
 * This avoids having to check for NULL before each call.
 *
 * Revision 1.8  2000/11/17 16:13:18  radhak
 * Fix some compile error and remove the POLL_FREE macro
 *
 * Revision 1.7  2000/11/17 14:28:05  radhak
 * Removed relMmapCredP.
 *
 * Revision 1.6  2000/11/10 22:45:51  radhak
 * Removed all mmap printks and added one more function to symbol table.
 *
 * Revision 1.4  2000/11/08 01:10:21  radhak
 * More linux mmap code
 *
 * Revision 1.3  2000/11/06 19:56:15  gjertsen
 * Linux code cleanup and put in build safeguards.
 *
 * Revision 1.2  2000/11/02 19:46:28  gjertsen
 * Linux code split. Pull out NBD stuff.
 *
 */

#include <Shark-gpl.h>
#include <arch-gpl.h>

#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/module.h>
#include <asm/pgalloc.h>
#include <linux/mman.h>
#include <linux/file.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/delay.h>

#include <cxiSystem.h>
#include <cxi2gpfs.h>
#include <cxiMmap.h>
#include <linux2gpfs.h>
#include <cxi2gpfs.h>
#include <Trace.h>

extern int nanosleep (__const struct timespec *__requested_time,
                      struct timespec *__remaining);
#define INTPAGER 10
#define Page_Skip(page)     test_bit(PG_skip, &(page)->flags)
#define ClearPageSkip(page) clear_bit(PG_skip, &(page)->flags)


#define MAX_PAGEQUE_ENTRIES 256
cxibuf_t Page_queue[MAX_PAGEQUE_ENTRIES];
cxibuf_t *PageQueueFreeP;

/* Head of list of bufs to be processed.  Bufs are placed on this list by
   the strategy routine (which runs at interrupt level) and removed by
   pager kprocs.  This list is circular and doubly-linked using av_forw and
   av_back. */
cxibuf_t PageRunQueue;

/* A lock to control access to the list of bufs to be processed.  Since the
   list is modified from interrupt level, disable_lock() must be used. */
Simple_lock pagerLock;

/* True if the pager kprocs have been started */
Boolean kprocsRunning = false;

/* If this flag is true, then all pagein or pageout requests will be returned
   immediately with an error. */
Boolean mmapAbort = false;

/* Event word which pager kprocs will wait on */
cxiWaitEvent_t pagerEventWord;

/* Event word to wait for free page queue entry */
cxiWaitEvent_t FreeEventWord;
int waitingForFreeEntries = 0;

/* LockWord and EventWord to serialize nopage() and mmapFlushseg() */
cxiBlockingMutex_t mmapFlushLockWord;
cxiWaitEvent_t mmapFlushEventWord;

#if LINUX_KERNEL_VERSION < 2041800
void 
flush_dirty_pages(struct inode * inode)
{
  struct list_head *tmp;
  struct page *page;
  int error;
  spinlock_t *page_hash_lock;
  struct address_space *mapping = inode->i_mapping;

try_again:
  TRACE0(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_DIRTY_PAGES_1,
         "flush_dirty_pages: getting address mapping lock\n");
  spin_lock(&mapping->page_lock);

  while ( !list_empty(&mapping->dirty_pages) )
  {
    tmp = mapping->dirty_pages.next;
    page = list_entry(tmp, struct page, list);

    TRACE4(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_DIRTY_PAGES_2,
           "flush_dirty_pages: page 0x%0X count %d index %d flags 0x%08X\n",
           page, page_count(page), page->index, page->flags);

    page_hash_lock = PAGECACHE_LOCK(page);
    if (!spin_trylock(page_hash_lock)) 
    {
      spin_unlock(&mapping->page_lock);
      goto try_again;
    }

    /* Move this page from dirty list to locked list */
    list_del(&page->list);
    list_add(&page->list, &mapping->locked_pages);
    if (!PageDirty(page)) 
    {
      spin_unlock(page_hash_lock);
      continue;
    }

    page_cache_get(page);
    spin_unlock(page_hash_lock);
    spin_unlock(&mapping->page_lock);

    if (PageLocked(page))
    {
      spin_lock(&mapping->page_lock);
      continue;
    }

    lock_page(page);
    if (PageDirty(page)) 
    {
      ClearPageDirty(page);
      if (page->mapping->a_ops->writepage)
      {
#if LINUX_KERNEL_VERSION >= 2040000
        error = page->mapping->a_ops->writepage(page);
        wait_on_page(page);
#else
        error = page->mapping->a_ops->writepage(vma->vm_file,page);
#endif
      }
    } 
    else
      UnlockPage(page);

    page_cache_release(page);
    spin_lock(&mapping->page_lock);
  }
  TRACE0(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_DIRTY_PAGES_6,
         "flush_dirty_pages:exit\n");
  spin_unlock(&mapping->page_lock);
}
#endif /* < 2041800 */

/* dump page contents
 * flag = 0 ==> after read from disk
 *        1 ==> write
 *        2 ==> before read from disk ( original contents of page)
 */
void dump_page(struct vm_area_struct *vma,struct page *page,int flag)  
{
#ifdef TRACE_IO_DATA
  int trcbuf[20];
  unsigned long kaddr;
  kaddr = (unsigned long)kmap(page);
  memcpy(trcbuf, (void *)kaddr, sizeof(trcbuf));
  TRACE8(TRACE_VNODE, 6, TRCID_MMAP_DIRTY_PAGE_DUMP,
     "dump 0 %s page 0x%lX: vma 0x%08X page count %d,data %08X %08X %08X %08X\n",
     (flag == 2)?"read_b":(flag == 1)?"write":"read",
     page,vma, page_count(page),
     CPUToBigEnd32(trcbuf[0]),
     CPUToBigEnd32(trcbuf[1]),
     CPUToBigEnd32(trcbuf[2]),
     CPUToBigEnd32(trcbuf[3]));
  TRACE8(TRACE_VNODE, 9, TRCID_MMAP_DIRTY_PAGE_DUMP_A,
     "dump 1 %s page 0x%lX: vma 0x%08X page count %d,data %08X %08X %08X %08X\n",
     (flag == 2)?"read_b":(flag == 1)?"write":"read",
     page,vma, page_count(page),
     CPUToBigEnd32(trcbuf[4]),
     CPUToBigEnd32(trcbuf[5]),
     CPUToBigEnd32(trcbuf[6]),
     CPUToBigEnd32(trcbuf[7]));
  TRACE8(TRACE_VNODE, 9, TRCID_MMAP_DIRTY_PAGE_DUMP_B,
     "dump 2 %s page 0x%lX: vma 0x%08X page count %d,data %08X %08X %08X %08X\n",
     (flag == 2)?"read_b":(flag == 1)?"write":"read",
     page,vma, page_count(page),
     CPUToBigEnd32(trcbuf[8]),
     CPUToBigEnd32(trcbuf[9]),
     CPUToBigEnd32(trcbuf[10]),
     CPUToBigEnd32(trcbuf[11]));
     kunmap(page);
#endif
}

void mmapFlushLock(void *cnPP,char *buf)
{
  cxiNode_t *cnP = (cxiNode_t *)cnPP;
  struct inode *iP = (struct inode *)cnP->osNodeP;
  pid_t pid = getpid();
  cxiBlockingMutexAcquire(&mmapFlushLockWord);
  while(cnP->mmapFlush)
  {
      TRACE3(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_LOCK_A,
           "%s is waiting for mmapFlushLock. mmapFlush %08X i_ino %d\n",buf,cnP->mmapFlush,iP->i_ino);
      cxiWaitEventWait(&mmapFlushEventWord, &mmapFlushLockWord, 0);
   }
  cnP->mmapFlush = true; 
  TRACE3(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_LOCK_B,
           "%s got mmapFlushLock. mmapFlush %08X i_ino %d\n",
	   buf,cnP->mmapFlush,iP->i_ino);
  cxiBlockingMutexRelease(&mmapFlushLockWord);
}

void mmapFlushUnLock(void *cnPP,char *buf)
{
  cxiNode_t *cnP = (cxiNode_t *)cnPP;
  struct inode *iP = (struct inode *)cnP->osNodeP;
  pid_t pid = getpid();
  cxiBlockingMutexAcquire(&mmapFlushLockWord);
  cnP->mmapFlush = false; 
  TRACE3(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_UNLOCK_A,
         "%s released mmapFlushLock. mmapFlush %08X i_ino %d\n",
	 buf,cnP->mmapFlush,iP->i_ino);
  cxiWaitEventSignal(&mmapFlushEventWord);
  cxiBlockingMutexRelease(&mmapFlushLockWord);
}

void mmapFlushUnLockNeeded(void *cnPP,char *buf)
{
  cxiNode_t *cnP = (cxiNode_t *)cnPP;
  struct inode *iP = (struct inode *)cnP->osNodeP;
  pid_t pid = getpid();
  cxiBlockingMutexAcquire(&mmapFlushLockWord);
  if (cnP->mmapFlush)
    cnP->mmapFlush = false; 
  TRACE3(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_UNLOCKNEED,
         "%s released mmapFlushLock. mmapFlush %08X i_ino %d\n",
	      buf,cnP->mmapFlush,iP->i_ino);
  cxiWaitEventSignal(&mmapFlushEventWord);
  cxiBlockingMutexRelease(&mmapFlushLockWord);
}

/* 
   gpfs_filemap_nopage() schedules pager kproc to wait until page table
   entry is created by do_no_page(). This function is called from pager 
   kproc. 
*/
void mmFlushWait(cxibuf_t *bufP)
{
   struct vm_area_struct *vma = (struct vm_area_struct *)bufP->vma;
   struct mm_struct * mm = vma->vm_mm;
   struct page *page = (struct page *)bufP->pageP;
   unsigned long offset =  (page->index) << PAGE_CACHE_SHIFT;
   unsigned long address =  offset + vma->vm_start;
   pgd_t * dir;
   pmd_t *pmd;
   pte_t *pte = NULL;
   Boolean try_again = false;
   char buf[] = "mmFlushWait after nopage()";
   cxiNode_t *cnP = bufP->b_vp;
   struct inode *iP = (struct inode *)cnP->osNodeP;
   int count = 0;
   int error = 0;
   pid_t pid = getpid();
   
pte_check:

#if defined(SUSE_LINUX) 
   /* in SuSE kernel, they kmap all the pte_offset() calls, when we're
      done with the pte we need to kunmap it */
   if(pte!=NULL) pte_kunmap(pte);
   pte = NULL;
#elif defined(REDHAT_AS_LINUX)
   if (pte != NULL) pte_unmap(pte);
   pte = NULL;
#endif

   dir = pgd_offset(vma->vm_mm, address);
   TRACE8(TRACE_VNODE, 6, TRCID_MMAP_FLUSH_WAIT_A,
      "mmFlushWait: vma %08X pageP %08X page index %d address %lX page count %d page falgs %08X loop_count %d i_ino %d\n",
      vma,bufP->pageP,page->index,address,page_count(page),page->flags,count,iP->i_ino);
   spin_lock(&mm->page_table_lock);
   TRACE0(TRACE_VNODE, 6, TRCID_MMAP_FLUSH_WAIT_B,
          "mmFlushWait: got page_table_lock\n");

   if (pgd_none(*dir))
   {
     TRACE1(TRACE_VNODE, 6, TRCID_MMAP_FLUSH_WAIT_B0,
       "mmFlushWait: pgd_none dir %lX)\n", dir);
     goto exit;
   }

   pmd = pmd_offset(dir,address);
   if (pmd_none(*pmd))
   {
     TRACE2(TRACE_VNODE, 6, TRCID_MMAP_FLUSH_WAIT_C,
       "mmFlushWait: pmd_none dir %lX  pmd %lX\n",dir, pmd);
     goto exit;
   }

#ifdef REDHAT_AS_LINUX
   pte = pte_offset_map(pmd,address);
#else
   pte = pte_offset(pmd,address);
#endif
   if (pte_present(*pte))
   {
     TRACE1(TRACE_VNODE, 6, TRCID_MMAP_FLUSH_WAIT_D,
         "mmFlushWait: pte_present pte %lX\n", (void *)pte);
     try_again = false;
   }
   else
    {
     TRACE1(TRACE_VNODE, 6, TRCID_MMAP_FLUSH_WAIT_E,
      "mmFlushWait: !pte_present pte %lX\n", pte);
     try_again = true;
    }
   
exit:
   spin_unlock(&mm->page_table_lock);
   TRACE0(TRACE_VNODE, 6, TRCID_MMAP_FLUSH_WAIT_F,
       "mmFlushWait: released page_table_lock\n");
   if (try_again && count < 5)
   {
      cxiSleep(1);
      count ++;
      TRACE1(TRACE_VNODE, 6, TRCID_MMAP_FLUSH_WAIT_A0,
         "mmFlushWait: try again loop count %d\n",count);
      goto pte_check;
   }

#if defined(SUSE_LINUX) 
   /* in SuSE kernel, they kmap all the pte_offset() calls, when we're
      done with the pte we need to kunmap it */
   if(pte!=NULL) pte_kunmap(pte);
#elif defined(REDHAT_AS_LINUX)
   if (pte != NULL) pte_unmap(pte);
#endif

   TRACE1(TRACE_VNODE, 6, TRCID_MMAP_FLUSH_WAIT_A1,
     "mmFlushWait: Releasing flush lock error %d \n",error);
   mmapFlushUnLock((void *)bufP->b_vp,buf);
}

static inline int invalidate_mmap_page(pte_t * ptep, 
                                       struct vm_area_struct *vma,
                                       struct page *pageP, 
                                       unsigned long address,
                                       enum MmflushOption mmfopt)
{
  pte_t pte = *ptep;
  struct page *page;
  int write_access = 0;
  int error = 0; 
  struct mm_struct * mm;

  TRACE4(TRACE_VNODE, 2, TRCID_MMAP_INV_ENTER,
         "invalidate_mmap_page: vma %08X address %08X mmfopt %d pageP %08X\n",
         vma, address, mmfopt, pageP);

  spin_lock(&vma->vm_mm->page_table_lock);
  if (!pte_present(pte))
  {
    TRACE0(TRACE_VNODE, 2, TRCID_MMAP_INV_PRESENT,
           "invalidate_mmap_page: pte not present\n");
    goto unlockpage;
  }
  if (pte_none(pte))
  {
    TRACE0(TRACE_VNODE, 2, TRCID_MMAP_INV_PTE_NONE,
           "invalidate_mmap_page: pte none\n");
    goto unlockpage;
  }
  
  if (vma->vm_flags & VM_WRITE)
     write_access++;
  page = pte_page(pte);
  if (page == pageP)
  {
    TRACE3(TRACE_VNODE, 2, TRCID_MMAP_INVALIDATE_PAGE,
           "Invalidating page 0x%0lX page count %d pte_dirty %d \n",
           pageP,page_count(pageP),pte_dirty(pte)?1:0);
    
    if (mmfopt == MmfInvalidate ||
        mmfopt == MmfDiscard ||
        mmfopt == MmfCacheInval)
    { 
#if defined(REDHAT_LINUX) && defined(MMAP_LINUX_PATCH) && \
	(LINUX_KERNEL_VERSION >= 2041800)
	  page_remove_rmap(page,ptep);
#endif

      pte_clear(ptep);

#if LINUX_KERNEL_VERSION >= 2040200
      FLUSH_TLB_PAGE(vma, address);
#endif

      /* Since this page is still in inode chache, page fault handler finds
         it from the cache and reads from disk only if the page is not 
         uptodate. Clear page uptodate bit so that next time when it is 
         referenced, page fault handler reads it from disk.
      */
      if (Page_Uptodate(page) && write_access && (vma->vm_flags & VM_SHARED))
         ClearPageUptodate(page);
      
      if (page_count(page) < 2)
         BUG();
        page_cache_release(page);
    }
    if (pte_dirty(pte) && write_access && page->mapping->a_ops->writepage)
    {
       TRACE2(TRACE_VNODE, 2, TRCID_MMAP_DIRTY_PAGE,
              "invalidate_mmap_page: dirty page 0x%0lX page count %d\n",
              page, page_count(page));
        dump_page(vma,page,1);
        SetPageDirty(page);
    }
  }

  unlockpage:
     spin_unlock(&vma->vm_mm->page_table_lock);
  TRACE1(TRACE_VNODE, 2, TRCID_MMAP_INV_EXIT,
           "invalidate_mmap_page exit error %d\n",error);
  return error;
}

int
pagerKprocMainThreadP(void *iparms)
{
  /* Release all user resources, including the file table */
  daemonize();
  strcpy(current->comm, "mmkproc");

  return gpfs_ops.pagerKprocMainThread(iparms);
}

int RegisterPagerBackEnd()
{
  int i;

  TRACE0(TRACE_VNODE, 2, TRCID_MMAP_REG_ENTER,
         "RegisterPagerBackEnd enter\n");

  /* Initialize page queue entries.  When a page arrives for read or write
     (by readpage or writepage functions), the page information will be
     copied to a free queue entry and that entry will be added to the end
     of the PageRunQueue. */
  PageQueueFreeP = NULL;
  for (i = 0; i < MAX_PAGEQUE_ENTRIES; i++)
  {
    Page_queue[i].av_forw = PageQueueFreeP;
    PageQueueFreeP = &Page_queue[i];
    Page_queue[i].pageP = NULL;
    Page_queue[i].b_vp = NULL;
    Page_queue[i].vinfoP = NULL;
    Page_queue[i].b_baddr = NULL;
    Page_queue[i].b_flags = 0;
    Page_queue[i].b_blkno = 0;
  }
  PageRunQueue.av_forw = PageRunQueue.av_back = &PageRunQueue;
  return 0;
}

/* Module termination */
int UnregisterPagerBackEnd()
{
  int rc = 0;
  TRACE0(TRACE_VNODE, 2, TRCID_MMAP_UNREG_ENTER,
         "UnregisterPagerBackEnd enter\n");
  PageQueueFreeP = NULL;
  PageRunQueue.av_forw = PageRunQueue.av_back = &PageRunQueue;

  TRACE1(TRACE_VNODE, 2, TRCID_MMAP_UNREG_EXIT,
         "UnregisterPagerBackEnd exit: rc %d\n", rc);
  return rc;
}

gpfs_Kpid_t
Createpagrkproc(void *iparms)
{
  gpfs_Kpid_t pid = -1;
  int rc = 0;

  /* Create the process */
  pid = kernel_thread(pagerKprocMainThreadP, iparms,
		      CLONE_FS | CLONE_FILES | SIGCHLD);
  TRACE2(TRACE_VNODE, 2, TRCID_MMAP_CREATE_EXIT,
         "Createpagrkproc exit: pid %d rc %d\n", pid, rc);
  return pid;
}

Int64 getFilePos(cxibuf_t *bufP)
{
  struct page *pageP = (struct page *)bufP->pageP;
  Int64 pos = ((Int64) pageP->index) << PAGE_CACHE_SHIFT;
  TRACE1(TRACE_VNODE, 2, TRCID_MMAP_FILEPOS_ENTER,
         "getFilePos: pos 0x%llX\n", pos);
  return pos;
}

/* This routine is called by the virtual memory manager when a page fault
   occurs on a mapped file.  All it does is place the paging request into a
   queue and wake up a pager kproc to process it. */
int gpfsRWpage(struct gpfsVfsData_t *privVfsP, cxiNode_t *cnP,
               struct MMFSVInfo *vinfoP,
               caddr_t kaddr, void * page, long write_req, void *vma)
{
  int rc = 0;
  cxiIntrState_t oldIntLvl;
  cxibuf_t *bufP;
  struct page *pageP = (struct page *)page;

  TRACE5(TRACE_VNODE, 2, TRCID_MMAP_RW_ENTER,
         "gpfsRWpage: %08x %08x %08x kaddr %08x req_type %s\n",
         pageP, PAGE_CACHE_SIZE, pageP->index, kaddr,
         write_req ? "write" : "read");

  /* Remember that there were paging requests under the given instance */
  if (vinfoP != NULL)
    ((cxiVinfo_t *)vinfoP)->rwPageDone = true;

  /* Lock out pager kprocs while we modify the buf list */
  oldIntLvl = disable_lock(INTPAGER, &pagerLock);

  /* Wait for a free request buf */
  while (PageQueueFreeP == NULL && kprocsRunning && !mmapAbort)
  {
    TRACE0(TRACE_VNODE, 2, TRCID_MMAP_WAIT_ENTER,
           "gpfsRWpage: waiting for request buf\n");
    waitingForFreeEntries++;
    e_sleep_thread(&FreeEventWord, &pagerLock, LOCK_HANDLER);
    waitingForFreeEntries--;
    DBGASSERT(waitingForFreeEntries >= 0);
  }
  
  /* Check that the kprocs are running and operation has not been aborted. */
  if (!kprocsRunning || mmapAbort)
  {
    /* The kprocs are not running or mapping has been aborted.  Either the
       daemon has not yet been started or it has been terminated.  Return
       all requests with an error. */
    rc = ENXIO;
  }
  else
  {
    /* Get a request buf from the free list */
    bufP = PageQueueFreeP;
    DBGASSERT(bufP != NULL);
    PageQueueFreeP = bufP->av_forw;

    bufP->b_flags = write_req;
    bufP->pageP = pageP;
    bufP->b_baddr = kaddr;
    bufP->b_bcount = PAGE_SIZE;
    bufP->b_vp = cnP;
    bufP->vinfoP = vinfoP;
    bufP->privVfsP = privVfsP;
    bufP->b_blkno = (PAGE_SIZE/CXI_DEV_BSIZE)*pageP->index;
    bufP->vma = vma;

    /* Add this request to the tail of the page run queue */
    bufP->av_forw = &PageRunQueue;
    bufP->av_back = PageRunQueue.av_back;
    PageRunQueue.av_back->av_forw = bufP;
    PageRunQueue.av_back = bufP;

    TRACE2(TRACE_VNODE, 2, TRCID_MMAP_RW_BUF,
        "gpfsRWpage: pageP 0x%lX. Queued buf 0x%lX to PageRunQueue\n",
        pageP, bufP);

    /* Wake up a pager kproc to process the request */
    cxiWaitEventWakeup(&pagerEventWord);
  }

  unlock_enable(oldIntLvl, &pagerLock);

  TRACE1(TRACE_VNODE, 2, TRCID_MMAP_RW_EXIT,
         "gpfsRWpage: returning with rc %d\n", rc);
  return rc;
}

void
IoDone(cxibuf_t *bufP)
{
  cxiIntrState_t oldIntLvl;
  cxiNode_t *cnP;
  pid_t pid = getpid();
  struct page *pageP = (struct page *)bufP->pageP;
  struct vm_area_struct *vma = (struct vm_area_struct *)bufP->vma;

  TRACE5(TRACE_VNODE, 2, TRCID_MMAP_IO_ENTER,
         "IoDone enter: pageP %08X, BufP->b_flags 0x%X count = %d "
         "page flags %08X index %d\n",
         pageP, bufP->b_flags, page_count(pageP), pageP->flags,pageP->index);

  /* error in read or write operation */
  if ((bufP->b_flags & B_ERROR) != 0)
  {
    ClearPageUptodate(pageP);
    /* This is the case when cancelBuf (close operation) is called immediately 
     * after nopage queues buffer to start a pagerkproc for waiting for pte 
     * entry creation. But nopage gets mmap flush lock before starting pager
     * kproc. We need to release mmap flush lock when called from cancelBuf
     */

    if ((bufP->b_flags & B_WAIT_FOR_PTE) != 0)
       mmapFlushUnLockNeeded(bufP->b_vp,"IoDone Error");
  }

  /* page table entry created. */
  if ((bufP->b_flags & B_WAIT_FOR_PTE) != 0)
    goto free_buf;

  /* read finished. set page flags and cleare invalidate bit */
  if ((bufP->b_flags & B_READ) != 0)
  {
    ClearPageSkip(pageP);
    SetPageUptodate(pageP);
    cnP = bufP->b_vp;
  }
  else
    ClearPageDirty(pageP);

  kunmap(pageP);
  UnlockPage(pageP);
  put_page(pageP);

free_buf:
  TRACE3(TRACE_VNODE, 2, TRCID_MMAP_IO_EXIT,
         "IoDone exit: pageP %08X, count = %d page flags %08X\n",
         pageP, page_count(pageP), pageP->flags);

  /* gpfs writepage synchronous operation and doesn't run by pager kproc. 
   * Also, no need to remove bufP from the pagequeue since it is not in
   * the queue.
   */
  if ((bufP->b_flags & (B_READ | B_WAIT_FOR_PTE)) != 0)
  {
    oldIntLvl = disable_lock(INTPAGER, &pagerLock);
    bufP->av_forw = PageQueueFreeP;
    PageQueueFreeP = bufP;

    /* wake up threads waiting for free entries */
    if (waitingForFreeEntries > 0)
      cxiWaitEventWakeup(&FreeEventWord);
    unlock_enable(oldIntLvl, &pagerLock);
  }
}

/* Wake up the kproc master thread */
void wakeKprocMaster(cxiPid_t pid)
{
  kill_proc(pid, SIGUSR1, 1);
}

cxibuf_t *
getHeadQueue()
{
  return &PageRunQueue;
}

void
getVp(void *gnP, void **vP, struct gpfsVfsData_t  **privVfsP)
{
  cxiNode_t *cP = (cxiNode_t *)gnP;
  struct inode *iP = (struct inode *)cP->osNodeP;
  *privVfsP = VP_TO_PVP(iP);
  *vP = cP->osNodeP;
}

void flush_vma(cxiNode_t *gnP, struct vm_area_struct *vma,
               unsigned long address, unsigned long end,
               enum MmflushOption mmfopt)
{
  pgd_t * dir;
  pgd_t *pgd; pmd_t *pmd; pte_t *pte = NULL;
  pte_t entry;
  int error = 0;
  struct page *page;
  unsigned int flags = 0;
  unsigned long start_address = address;
  size_t size = end - address;
  unsigned long  index;
  struct file *file = vma->vm_file;
  struct inode *inode = file->f_dentry->d_inode;
  struct address_space *mapping = inode->i_mapping;
  Boolean need_unlock = false;
   pid_t pid = getpid();

  TRACE3(TRACE_VNODE, 2, TRCID_FLUSH_VMA_ENTER,
         "flush_vma: vma %lX address %lX size %d\n", vma, address, size);

  if (vma->vm_file == NULL)
  {
    error = 1;
    goto exit1;
  }

  get_file(vma->vm_file);

  if (address >= end)
  {
    TRACE0(TRACE_VNODE, 2, TRCID_FLUSH_VMA_BAD_ADDR,
           "flush_vma: bad address\n");
    error = 2;
    goto exit;
  }
  
  /* find page table directory entry */
  spin_lock(&vma->vm_mm->page_table_lock); 
  flush_cache_range(vma->vm_mm, address, end);
  TRACE1(TRACE_VNODE, 6, TRCID_FLUSH_VMA_PGT_LOCK,
           "flush_vma: got vma %08X page_table_lock\n",vma);
   need_unlock = true;
 
   dir = pgd_offset(vma->vm_mm, address);
   if (pgd_none(*dir))
   {
       TRACE0(TRACE_VNODE, 6, TRCID_FLUSH_VMA_BAD_PGD, 
           "flush_vma: pgd none\n");
       goto exit;
   }
    
  do
  {
    index = ((address - vma->vm_start) >> PAGE_CACHE_SHIFT) + vma->vm_pgoff;
    TRACE4(TRACE_VNODE, 6, TRCID_PAGE_LOCK,
           "flush_vma: getting locked page. vma 0x%lX address 0x%lX end 0x%lX index %d\n",
            vma,address,end,index);
    pmd = pmd_offset(dir,address);
    if (pmd_none(*pmd))
    {
       TRACE0(TRACE_VNODE, 6, TRCID_FLUSH_VMA_BAD_PMD, 
               "flush_vma: pmd none\n");
       goto next_page;
    }
#ifdef REDHAT_AS_LINUX
    pte = pte_offset_map(pmd,address);
#else
    pte = pte_offset(pmd,address);
#endif
    if (!pte_present(*pte))
    {
        TRACE0(TRACE_VNODE, 6, TRCID_FLUSH_VMA_BAD_PTE, 
		      "flush_vma: pte not present\n");
        goto next_page;
    }
    if (pte_none(*pte))
    {
        TRACE0(TRACE_VNODE, 6, TRCID_FLUSH_VMA_NONE_PTE,
            "flush_vma: pte none\n");
        goto next_page;
     }

     entry = *pte;
     page = pte_page(entry);
    
    /* make sure this page belongs to this mapping. */
    
    if (page == NULL)
      goto next_page; 

    if ( (page->mapping != mapping) || (page->index != index) )
      goto next_page;

    /* readpage must be still reading this page from disk */
    /* gpfs_i_readpage sets PageSkip flag to let invalidation function know */
    if (Page_Skip(page))
    {
      goto next_page;
    }

    page_cache_get(page);
    TRACE1(TRACE_VNODE, 6, TRCID_FLUSH_VMA_PAGE_LOCK,
          "flush_vma: getting page %08X lock\n",page);
    /* do not hold spin lock while getting blocking lock */
    spin_unlock(&vma->vm_mm->page_table_lock);
    lock_page(page);
    error |= invalidate_mmap_page(pte, vma, page,address,mmfopt);
    if (PageDirty(page))
    {
      error = page->mapping->a_ops->writepage(page);
    }
    else
      UnlockPage(page); 

#if LINUX_KERNEL_VERSION >= 2041000
    page_cache_release(page);
#else
    page_cache_free(page);
#endif
    spin_lock(&vma->vm_mm->page_table_lock); 

    next_page:

#if defined(SUSE_LINUX)
	/* in SuSE kernel, they kmap all the pte_offset() calls, when we're
       done with the pte we need to kunmap it */
    if(pte!=NULL) pte_kunmap(pte);
#elif defined(REDHAT_AS_LINUX)
   if (pte != NULL) pte_unmap(pte);
#endif

    address += PAGE_SIZE;

  } while (address && (address < end));

exit:
  if (need_unlock)
  { 
    spin_unlock(&vma->vm_mm->page_table_lock);
    TRACE1(TRACE_VNODE, 6, TRCID_FLUSH_VMA_PGT_UNLOCK,
           "flush_vma: releasing vma %08X page_table_lock\n",vma);
  }

  fput(vma->vm_file);
exit1:
  TRACE1(TRACE_VNODE, 2, TRCID_FLUSH_VMA_EXIT,
         "flush_vma:error %d\n", error);
}

void scan_vma(cxiNode_t *gnP, struct vm_area_struct *vma,  
               unsigned long address,
               enum MmflushOption mmfopt,struct page **pageP)
{
  pgd_t * dir;
  pgd_t *pgd; pmd_t *pmd; pte_t *pte = NULL;
  pte_t entry;
  int error = 0;
  struct page *page = NULL;
  unsigned int flags = 0;
  size_t size = 0;
  unsigned long  index;
  struct file *file = vma->vm_file;
  struct inode *inode = file->f_dentry->d_inode;
  struct address_space *mapping = inode->i_mapping;
  int code = 0;
  Boolean need_unlock = false;
  pid_t pid = getpid();

  if (vma->vm_file == NULL)
  {
    code = 100;
    goto exit2;
  }

  get_file(vma->vm_file);

  flush_cache_range(vma->vm_mm, vma->vma_start,vma->vma_end);
  spin_lock(&vma->vm_mm->page_table_lock); 
  TRACE1(TRACE_VNODE, 6, TRCID_SCAN_VMA_PGT_LOCK,
           "scan_vma: got vma %08X page_table_lock\n",vma);

  need_unlock = true;
  /* find page table directory entry */
   dir = pgd_offset(vma->vm_mm, address);
   if (pgd_none(*dir))
   {
	   code = 1;
	   goto exit;
   }
   index = ((address - vma->vm_start) >> PAGE_CACHE_SHIFT) + vma->vm_pgoff;
   TRACE4(TRACE_VNODE, 6, TRCID_SCAN_VMA_1,
       "scan_vma: start 0x%lX end 0x%lX vm_pgoff %d PAGE_CACHE_SHIFT %d\n",
           vma->vm_start,vma->vm_end,vma->vm_pgoff,PAGE_CACHE_SHIFT);

   TRACE3(TRACE_VNODE, 6, TRCID_SCAN_VMA_2,
      "scan_vma: getting locked page. vma 0x%lX address 0x%lX index %d\n",
            vma,address,index);

    pmd = pmd_offset(dir,address);
    if (pmd_none(*pmd))
    {
       TRACE0(TRACE_VNODE, 6, TRCID_SCAN_VMA_ERR1,
	          "scan_vma: pmd_none\n");
	    code = 2;
	    goto exit;
    }
#ifdef REDHAT_AS_LINUX
    pte = pte_offset_map(pmd,address);
#else
    pte = pte_offset(pmd,address);
#endif
    if (!pte_present(*pte))
    {
      TRACE0(TRACE_VNODE, 6, TRCID_SCAN_VMA_ERR2,
             "scan_vma:pte not present\n");
      code = 3;
      goto exit;
    }
    if (pte_none(*pte))
    {
      TRACE0(TRACE_VNODE, 6, TRCID_SCAN_VMA_ERR3,
             "scan_vma:pte none\n");
      code = 4;
      goto exit;
    }
    entry = *pte;
    page = pte_page(entry);

   TRACE2(TRACE_VNODE, 6, TRCID_SCAN_VMA_3,
          "scan_vma: page %lX *pageP %lX\n",page,*pageP);
    
    /* make sure this page belongs to this mapping. */
    
    if (page == NULL)
    {
      code = 5;
      goto exit; 
    }

    if ( (page->mapping != mapping) || (page->index != index) )
    {
      code = 6;
      goto exit;
    }

    if (*pageP && (page != *pageP) )
    {
      code = 7;
      goto exit;
    }
    else
    {
      if ( *pageP == NULL)
      {
        if (Page_Skip(page))
        {
          code = 8;
          goto exit;
        }
        *pageP = page;
        page_cache_get(page);
	    spin_unlock(&vma->vm_mm->page_table_lock);

        TRACE1(TRACE_VNODE, 6, TRCID_SCAN_VMA_6,
               "scan_vma: getting page %08X lock\n",page);
        lock_page(page);

        TRACE2(TRACE_VNODE, 6, TRCID_SCAN_VMA_7,
               "scan_vma: got page %08X lock. page count %d\n",
                page,page_count(page));
      }
      else
        spin_unlock(&vma->vm_mm->page_table_lock);
    }
    need_unlock = false;
    error |= invalidate_mmap_page(pte, vma, page,address,mmfopt);

    TRACE2(TRACE_VNODE, 6, TRCID_SCAN_VMA_4,
           "invalidated page %08X page count  %d\n",
            page,page_count(page));

exit:
  fput(vma->vm_file);

#if defined(SUSE_LINUX)
	/* in SuSE kernel, they kmap all the pte_offset() calls, when we're
       done with the pte we need to kunmap it */
    if(pte!=NULL) pte_kunmap(pte);
#elif defined(REDHAT_AS_LINUX)
   if (pte != NULL) pte_unmap(pte);
#endif

  if (need_unlock)
    spin_unlock(&vma->vm_mm->page_table_lock);
exit2:
  TRACE2(TRACE_VNODE, 6, TRCID_SCAN_VMA_5,
     "scan_vma:error %d code %d\n", error,code);
}

void flushVma(cxiNode_t *gnP, struct vm_area_struct *shared_vma,
               UInt64 start, UInt64 end,
               enum MmflushOption mmfopt)
{
  unsigned long offset, start_address, end_address,address;
  struct mm_struct *mm;
  Boolean lockheld = false;
  struct page *pageP = NULL;
  int error;
  struct vm_area_struct *vma;
  pid_t pid = getpid();

  struct inode *inodeP = (struct inode *) gnP->osNodeP;
  address = start;
  do
  {
    vma = shared_vma;
    if (vma == NULL)
    {
	   return;
    }
    do
    {
      TRACE2(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_1,
           "flushVma: vma = %08X, i_no = %d\n",
           vma, vma->vm_file->f_dentry->d_inode->i_ino);
      TRACE5(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_2,
           "flushVma: vma_start = %08X, vma_end = %08X address %08X start %08X end %08X\n",
           vma->vm_start, vma->vm_end,address,start,end);

      /* It should never happen */
      if (inodeP->i_ino != vma->vm_file->f_dentry->d_inode->i_ino)
      continue;

      /* No need to flush if range is outside the limits of what can be
       mapped */

      offset = vma->vm_pgoff << PAGE_SHIFT;
      if (start > (vma->vm_end - vma->vm_start)  || end < offset)
        continue;

      if (address < offset)
        continue;
      start_address = vma->vm_start + address;
     
      TRACE2(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_3,
           "flushVma: start_address %lX end_address %lX\n",
           start_address, end_address);

      /* flush this vma */
      if (vma->vm_flags & VM_DENYWRITE && vma->vm_flags & VM_WRITE)
      {
        TRACE1(TRACE_VNODE, 3, TRCID_MMAP_FLUSH_4,
             "do not flush vma 0x%lX\n",vma);
        continue;
      }
      if (mmfopt == MmfKeep)
        filemap_sync(vma,vma->vm_start,(vma->vm_end - vma->vm_start),0);
      else
        scan_vma(gnP, vma, start_address,  mmfopt,&pageP);
      TRACE1(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_7,
            "finished vma %08X\n",vma);
    } while ((vma = vma->vm_next_share) != NULL);

    TRACE0(TRACE_VNODE, 6, TRCID_MMAP_FLUSH_7A,
          "finished all vmas\n");
    if (pageP)
    {
       if (PageDirty(pageP))
       {
         TRACE2(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_8,
                "flushVma: flushing page 0x%08X page count %d\n",
                 pageP,page_count(pageP));
          error = pageP->mapping->a_ops->writepage(pageP);
       }
       else
          UnlockPage(pageP);

       TRACE2(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_9,
              "flushVma: finished page 0x%08X page count %d\n",
               pageP,page_count(pageP));
       page_cache_release(pageP);
    }
next_page:
    address += PAGE_SIZE;
    pageP = NULL;

    TRACE1(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_5,
        "flushVma:next address %lX\n",address);
  } while (address && (address < end));

  TRACE0(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_6,
  "flushVma:finished all address\n");
}
void flushVma_keep(cxiNode_t *gnP, struct vm_area_struct *vma,
               UInt64 start, UInt64 end,
               enum MmflushOption mmfopt)
{
  unsigned long offset, start_address, end_address;
  struct mm_struct *mm;
  Boolean lockheld = false;

  struct inode *inodeP = (struct inode *) gnP->osNodeP;
  if (vma == NULL)
	return;

  do
  {
    TRACE2(TRACE_VNODE, 2, TRCID_MMAP_FLUSHVMA_1,
           "flushVma: vma = %08X, i_no = %d\n",
           vma, vma->vm_file->f_dentry->d_inode->i_ino);
    TRACE2(TRACE_VNODE, 2, TRCID_MMAP_FLUSHVMA_2,
           "flushVma: vma_start = %08X, vma_end = %08X\n",
           vma->vm_start, vma->vm_end);

    /* It should never happen */
    if (inodeP->i_ino != vma->vm_file->f_dentry->d_inode->i_ino)
      continue;

    /* No need to flush if range is outside the limits of what can be
       mapped */

    offset = vma->vm_pgoff << PAGE_SHIFT;
    if (start >= (vma->vm_end - vma->vm_start)  || end <= offset)
      continue;

    if (start <= offset)
      start_address = vma->vm_start;
    else
      start_address = vma->vm_start + start ;

    if (end >= (vma->vm_end - vma->vm_start) )
      end_address = vma->vm_end;
    else
      end_address = vma->vm_start + end;

    TRACE2(TRACE_VNODE, 2, TRCID_MMAP_FLUSHVMA_3,
           "flushVma: start_address %lX end_address %lX\n",
           start_address, end_address);

    /* flush this vma */
    if (vma->vm_flags & VM_DENYWRITE && vma->vm_flags & VM_WRITE)
    {
      TRACE1(TRACE_VNODE, 3, TRCID_MMAP_FLUSHVMA_4,
             "do not flush vma 0x%lX\n",vma);
      continue;
    }
    if (mmfopt == MmfKeep)
      filemap_sync(vma,vma->vm_start,(vma->vm_end - vma->vm_start),0);
    else
      flush_vma(gnP, vma, start_address, end_address, mmfopt);
  } while ((vma = vma->vm_next_share) != NULL);
}

/* Flush Options:
       MmfKeep       - Leave pages as they are
       MmfProtect    - Change page protection to read-only
       MmfCancel     - Cancel pending requests only (no flush or invalidate)
       MmfTerminate  - Cancel pending requests
       MmfInvalidate - Invalidate pages
       MmfDiscard    - Invalidate pages without flushing them
       MmfCacheInval - Invalidate pages even if file is no longer mapped
 */

MMFlushRange(void *gnPP, UInt64 start, UInt64 end, enum MmflushOption mmfopt)
{
  int rc = 0;
  struct vm_area_struct *vma, *shared_vma = NULL;
  cxiNode_t *gnP = (cxiNode_t *)gnPP;
  struct inode *inodeP;
  struct address_space *mapping;
  char buf[] = "MMFlushRange()";
  pid_t pid = getpid();
  Boolean do_unlock = false;

  DBGASSERT(gnP != NULL);
  
  inodeP = (struct inode *) gnP->osNodeP;
  mapping = inodeP->i_mapping;

  TRACE4(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_ENTER,
         "MMFlushRange: gnP %08X inode number = %d, mmopt %d nrpages %d\n",
         gnP, inodeP->i_ino, mmfopt,inodeP->i_data.nrpages);
  TRACE2(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_ENTER_A,
         "MMFlushRange: start 0x%lX end 0x%lX\n",
         (start &PAGE_MASK),(end & PAGE_MASK));

 if (end >inodeP->i_size)
	 end = inodeP->i_size; 
 if (mmfopt == MmfInvalidate ||
     mmfopt == MmfDiscard ||
     mmfopt == MmfCacheInval)
 {
   mmapFlushLock((void *)gnP,buf);
   do_unlock = true;
 }

  spin_lock(&mapping->i_shared_lock);

# if LINUX_KERNEL_VERSION >= 2040000
  shared_vma = mapping->i_mmap_shared;
  if (shared_vma) 
    if (mmfopt == MmfKeep)
       flushVma_keep(gnP,shared_vma,start,end,mmfopt);
    else
       flushVma(gnP,shared_vma,start,end,mmfopt);
#else
    vma = mapping->i_mmap;
    if (vma)
    if (mmfopt == MmfKeep)
       flushVma_keep(gnP,vma,start,end,mmfopt);
    else
       flushVma(gnP,vma,start,end,mmfopt);
#endif

  spin_unlock(&mapping->i_shared_lock);

flush_dirty:  
  if (mmfopt != MmfKeep)
  {
 
   TRACE2(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_INVAL1,
           "Invalidating inode pages of inode %d nrpages %d\n",
            inodeP->i_ino,inodeP->i_data.nrpages);
   /* First flush all dirty pages if any */
   if (mmfopt == MmfTerminate)
   {
      /* MmfInvalidate already flushed dirty pages while invalidating
         page table entries there is no need to do it again
         for MmfInvalidate option.
      */

#if LINUX_KERNEL_VERSION < 2041800
   flush_dirty_pages(inodeP);
#else
	/* flush all dirty mmaped pages */
   filemap_fdatasync(inodeP->i_mapping);
   /* Now since writepage is synchronous, no need to wait until IO is
    * finished
   */
#endif
   }
    /* invalidate pages from page cache */
    invalidate_inode_pages(inodeP);
    TRACE2(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_INVAL2,
           "Invalidated inode pages of inode %d nrpages %d\n",
            inodeP->i_ino,inodeP->i_data.nrpages);
  }

 if (do_unlock)
    mmapFlushUnLock((void *)gnP,buf);

exit:
  TRACE1(TRACE_VNODE, 2, TRCID_MMAP_FLUSH_EXIT,
         "MMFlushRange exit: rc %d\n", rc);
  return rc;
}
